/*
 * QEMU model of the CRL_APB APB control registers for clock controller. The rstctrl_lpd willbe added to this as well
 *
 * Copyright (c) 2014 Xilinx Inc.
 *
 * Autogenerated by xregqemu.py 2014-03-20.
 * Written by Edgar E. Iglesias
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "qemu/osdep.h"
#include "hw/sysbus.h"
#include "hw/register-dep.h"
#include "qemu/bitops.h"
#include "qemu/log.h"
#include "qemu/config-file.h"
#include "sysemu/sysemu.h"

#include "hw/fdt_generic_util.h"

#ifndef XILINX_CRL_APB_ERR_DEBUG
#define XILINX_CRL_APB_ERR_DEBUG 0
#endif

#define TYPE_XILINX_CRL_APB "xlnx.zynqmp-crl"

#define XILINX_CRL_APB(obj) \
     OBJECT_CHECK(CRL_APB, (obj), TYPE_XILINX_CRL_APB)

DEP_REG32(ERR_CTRL, 0x0)
    DEP_FIELD(ERR_CTRL, SLVERR_ENABLE, 1, 0)
DEP_REG32(IR_STATUS, 0x4)
    DEP_FIELD(IR_STATUS, ADDR_DECODE_ERR, 1, 0)
DEP_REG32(IR_MASK, 0x8)
    DEP_FIELD(IR_MASK, ADDR_DECODE_ERR, 1, 0)
DEP_REG32(IR_ENABLE, 0xc)
    DEP_FIELD(IR_ENABLE, ADDR_DECODE_ERR, 1, 0)
DEP_REG32(IR_DISABLE, 0x10)
    DEP_FIELD(IR_DISABLE, ADDR_DECODE_ERR, 1, 0)
DEP_REG32(CRL_WPROT, 0x1C)
    DEP_FIELD(CRL_WPROT, ACTIVE, 1, 0)
DEP_REG32(IOPLL_CTRL, 0x20)
    DEP_FIELD(IOPLL_CTRL, POST_SRC, 3, 24)
    DEP_FIELD(IOPLL_CTRL, PRE_SRC, 3, 20)
    DEP_FIELD(IOPLL_CTRL, DIV2, 1, 16)
    DEP_FIELD(IOPLL_CTRL, FBDIV, 7, 8)
    DEP_FIELD(IOPLL_CTRL, BYPASS, 1, 3)
    DEP_FIELD(IOPLL_CTRL, RESET, 1, 0)
DEP_REG32(IOPLL_CFG, 0x24)
    DEP_FIELD(IOPLL_CFG, LOCK_DLY, 7, 25)
    DEP_FIELD(IOPLL_CFG, LOCK_CNT, 10, 13)
    DEP_FIELD(IOPLL_CFG, LFHF, 2, 10)
    DEP_FIELD(IOPLL_CFG, CP, 4, 5)
    DEP_FIELD(IOPLL_CFG, RES, 4, 0)
DEP_REG32(IOPLL_FRAC_CFG, 0x28)
    DEP_FIELD(IOPLL_FRAC_CFG, ENABLED, 1, 31)
    DEP_FIELD(IOPLL_FRAC_CFG, SEED, 3, 22)
    DEP_FIELD(IOPLL_FRAC_CFG, ALGRTHM, 1, 19)
    DEP_FIELD(IOPLL_FRAC_CFG, ORDER, 1, 18)
    DEP_FIELD(IOPLL_FRAC_CFG, DATA, 16, 0)
DEP_REG32(RPLL_CTRL, 0x30)
    DEP_FIELD(RPLL_CTRL, POST_SRC, 3, 24)
    DEP_FIELD(RPLL_CTRL, PRE_SRC, 3, 20)
    DEP_FIELD(RPLL_CTRL, DIV2, 1, 16)
    DEP_FIELD(RPLL_CTRL, FBDIV, 7, 8)
    DEP_FIELD(RPLL_CTRL, BYPASS, 1, 3)
    DEP_FIELD(RPLL_CTRL, RESET, 1, 0)
DEP_REG32(RPLL_CFG, 0x34)
    DEP_FIELD(RPLL_CFG, LOCK_DLY, 7, 25)
    DEP_FIELD(RPLL_CFG, LOCK_CNT, 10, 13)
    DEP_FIELD(RPLL_CFG, LFHF, 2, 10)
    DEP_FIELD(RPLL_CFG, CP, 4, 5)
    DEP_FIELD(RPLL_CFG, RES, 4, 0)
DEP_REG32(RPLL_FRAC_CFG, 0x38)
    DEP_FIELD(RPLL_FRAC_CFG, ENABLED, 1, 31)
    DEP_FIELD(RPLL_FRAC_CFG, SEED, 3, 22)
    DEP_FIELD(RPLL_FRAC_CFG, ALGRTHM, 1, 19)
    DEP_FIELD(RPLL_FRAC_CFG, ORDER, 1, 18)
    DEP_FIELD(RPLL_FRAC_CFG, DATA, 16, 0)
DEP_REG32(PLL_STATUS, 0x40)
    DEP_FIELD(PLL_STATUS, RPLL_STABLE, 1, 4)
    DEP_FIELD(PLL_STATUS, IOPLL_STABLE, 1, 3)
    DEP_FIELD(PLL_STATUS, RPLL_LOCK, 1, 1)
    DEP_FIELD(PLL_STATUS, IOPLL_LOCK, 1, 0)
DEP_REG32(IOPLL_TO_FPD_CTRL, 0x44)
    DEP_FIELD(IOPLL_TO_FPD_CTRL, DIVISOR0, 6, 8)
DEP_REG32(RPLL_TO_FPD_CTRL, 0x48)
    DEP_FIELD(RPLL_TO_FPD_CTRL, DIVISOR0, 6, 8)
DEP_REG32(USB3_DUAL_REF_CTRL, 0x4c)
    DEP_FIELD(USB3_DUAL_REF_CTRL, CLKACT, 1, 25)
    DEP_FIELD(USB3_DUAL_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(USB3_DUAL_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(USB3_DUAL_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(GEM0_REF_CTRL, 0x50)
    DEP_FIELD(GEM0_REF_CTRL, RX_CLKACT, 1, 26)
    DEP_FIELD(GEM0_REF_CTRL, CLKACT, 1, 25)
    DEP_FIELD(GEM0_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(GEM0_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(GEM0_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(GEM1_REF_CTRL, 0x54)
    DEP_FIELD(GEM1_REF_CTRL, RX_CLKACT, 1, 26)
    DEP_FIELD(GEM1_REF_CTRL, CLKACT, 1, 25)
    DEP_FIELD(GEM1_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(GEM1_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(GEM1_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(GEM2_REF_CTRL, 0x58)
    DEP_FIELD(GEM2_REF_CTRL, RX_CLKACT, 1, 26)
    DEP_FIELD(GEM2_REF_CTRL, CLKACT, 1, 25)
    DEP_FIELD(GEM2_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(GEM2_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(GEM2_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(GEM3_REF_CTRL, 0x5c)
    DEP_FIELD(GEM3_REF_CTRL, RX_CLKACT, 1, 26)
    DEP_FIELD(GEM3_REF_CTRL, CLKACT, 1, 25)
    DEP_FIELD(GEM3_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(GEM3_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(GEM3_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(USB0_BUS_REF_CTRL, 0x60)
    DEP_FIELD(USB0_BUS_REF_CTRL, CLKACT, 1, 25)
    DEP_FIELD(USB0_BUS_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(USB0_BUS_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(USB0_BUS_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(USB1_BUS_REF_CTRL, 0x64)
    DEP_FIELD(USB1_BUS_REF_CTRL, CLKACT, 1, 25)
    DEP_FIELD(USB1_BUS_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(USB1_BUS_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(USB1_BUS_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(QSPI_REF_CTRL, 0x68)
    DEP_FIELD(QSPI_REF_CTRL, CLKACT, 1, 24)
    DEP_FIELD(QSPI_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(QSPI_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(QSPI_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(SDIO0_REF_CTRL, 0x6c)
    DEP_FIELD(SDIO0_REF_CTRL, CLKACT, 1, 24)
    DEP_FIELD(SDIO0_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(SDIO0_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(SDIO0_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(SDIO1_REF_CTRL, 0x70)
    DEP_FIELD(SDIO1_REF_CTRL, CLKACT, 1, 24)
    DEP_FIELD(SDIO1_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(SDIO1_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(SDIO1_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(UART0_REF_CTRL, 0x74)
    DEP_FIELD(UART0_REF_CTRL, CLKACT, 1, 24)
    DEP_FIELD(UART0_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(UART0_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(UART0_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(UART1_REF_CTRL, 0x78)
    DEP_FIELD(UART1_REF_CTRL, CLKACT, 1, 24)
    DEP_FIELD(UART1_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(UART1_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(UART1_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(SPI0_REF_CTRL, 0x7c)
    DEP_FIELD(SPI0_REF_CTRL, CLKACT, 1, 24)
    DEP_FIELD(SPI0_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(SPI0_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(SPI0_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(SPI1_REF_CTRL, 0x80)
    DEP_FIELD(SPI1_REF_CTRL, CLKACT, 1, 24)
    DEP_FIELD(SPI1_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(SPI1_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(SPI1_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(CAN0_REF_CTRL, 0x84)
    DEP_FIELD(CAN0_REF_CTRL, CLKACT, 1, 24)
    DEP_FIELD(CAN0_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(CAN0_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(CAN0_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(CAN1_REF_CTRL, 0x88)
    DEP_FIELD(CAN1_REF_CTRL, CLKACT, 1, 24)
    DEP_FIELD(CAN1_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(CAN1_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(CAN1_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(CPU_R5_CTRL, 0x90)
    DEP_FIELD(CPU_R5_CTRL, CLKACT_CORE, 1, 25)
    DEP_FIELD(CPU_R5_CTRL, CLKACT, 1, 24)
    DEP_FIELD(CPU_R5_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(CPU_R5_CTRL, SRCSEL, 3, 0)
DEP_REG32(IOU_SWITCH_CTRL, 0x9c)
    DEP_FIELD(IOU_SWITCH_CTRL, CLKACT, 1, 24)
    DEP_FIELD(IOU_SWITCH_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(IOU_SWITCH_CTRL, SRCSEL, 3, 0)
DEP_REG32(CSU_PLL_CTRL, 0xa0)
    DEP_FIELD(CSU_PLL_CTRL, CLKACT, 1, 24)
    DEP_FIELD(CSU_PLL_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(CSU_PLL_CTRL, SRCSEL, 3, 0)
DEP_REG32(PCAP_CTRL, 0xa4)
    DEP_FIELD(PCAP_CTRL, CLKACT, 1, 24)
    DEP_FIELD(PCAP_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(PCAP_CTRL, SRCSEL, 3, 0)
DEP_REG32(LPD_SWITCH_CTRL, 0xa8)
    DEP_FIELD(LPD_SWITCH_CTRL, CLKACT, 1, 24)
    DEP_FIELD(LPD_SWITCH_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(LPD_SWITCH_CTRL, SRCSEL, 3, 0)
DEP_REG32(LPD_LSBUS_CTRL, 0xac)
    DEP_FIELD(LPD_LSBUS_CTRL, CLKACT, 1, 24)
    DEP_FIELD(LPD_LSBUS_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(LPD_LSBUS_CTRL, SRCSEL, 3, 0)
DEP_REG32(DBG_LPD_CTRL, 0xb0)
    DEP_FIELD(DBG_LPD_CTRL, CLKACT, 1, 24)
    DEP_FIELD(DBG_LPD_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(DBG_LPD_CTRL, SRCSEL, 3, 0)
DEP_REG32(NAND_REF_CTRL, 0xb4)
    DEP_FIELD(NAND_REF_CTRL, CLKACT, 1, 24)
    DEP_FIELD(NAND_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(NAND_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(NAND_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(ADMA_REF_CTRL, 0xb8)
    DEP_FIELD(ADMA_REF_CTRL, CLKACT, 1, 24)
    DEP_FIELD(ADMA_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(ADMA_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(PL0_REF_CTRL, 0xc0)
    DEP_FIELD(PL0_REF_CTRL, CLKACT, 1, 24)
    DEP_FIELD(PL0_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(PL0_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(PL0_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(PL1_REF_CTRL, 0xc4)
    DEP_FIELD(PL1_REF_CTRL, CLKACT, 1, 24)
    DEP_FIELD(PL1_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(PL1_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(PL1_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(PL2_REF_CTRL, 0xc8)
    DEP_FIELD(PL2_REF_CTRL, CLKACT, 1, 24)
    DEP_FIELD(PL2_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(PL2_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(PL2_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(PL3_REF_CTRL, 0xcc)
    DEP_FIELD(PL3_REF_CTRL, CLKACT, 1, 24)
    DEP_FIELD(PL3_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(PL3_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(PL3_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(PL0_THR_CTRL, 0xd0)
    DEP_FIELD(PL0_THR_CTRL, CURR_VAL, 16, 16)
    DEP_FIELD(PL0_THR_CTRL, RUNNING, 1, 15)
    DEP_FIELD(PL0_THR_CTRL, CPU_START, 1, 1)
    DEP_FIELD(PL0_THR_CTRL, CNT_RST, 1, 0)
DEP_REG32(PL0_THR_CNT, 0xd4)
    DEP_FIELD(PL0_THR_CNT, LAST_CNT, 16, 0)
DEP_REG32(PL1_THR_CTRL, 0xd8)
    DEP_FIELD(PL1_THR_CTRL, CURR_VAL, 16, 16)
    DEP_FIELD(PL1_THR_CTRL, RUNNING, 1, 15)
    DEP_FIELD(PL1_THR_CTRL, CPU_START, 1, 1)
    DEP_FIELD(PL1_THR_CTRL, CNT_RST, 1, 0)
DEP_REG32(PL1_THR_CNT, 0xdc)
    DEP_FIELD(PL1_THR_CNT, LAST_CNT, 16, 0)
DEP_REG32(PL2_THR_CTRL, 0xe0)
    DEP_FIELD(PL2_THR_CTRL, CURR_VAL, 16, 16)
    DEP_FIELD(PL2_THR_CTRL, RUNNING, 1, 15)
    DEP_FIELD(PL2_THR_CTRL, CPU_START, 1, 1)
    DEP_FIELD(PL2_THR_CTRL, CNT_RST, 1, 0)
DEP_REG32(PL2_THR_CNT, 0xe4)
    DEP_FIELD(PL2_THR_CNT, LAST_CNT, 16, 0)
DEP_REG32(PL3_THR_CTRL, 0xe8)
    DEP_FIELD(PL3_THR_CTRL, CURR_VAL, 16, 16)
    DEP_FIELD(PL3_THR_CTRL, RUNNING, 1, 15)
    DEP_FIELD(PL3_THR_CTRL, CPU_START, 1, 1)
    DEP_FIELD(PL3_THR_CTRL, CNT_RST, 1, 0)
DEP_REG32(PL3_THR_CNT, 0xfc)
    DEP_FIELD(PL3_THR_CNT, LAST_CNT, 16, 0)
DEP_REG32(GEM_TSU_REF_CTRL, 0x100)
    DEP_FIELD(GEM_TSU_REF_CTRL, CLKACT, 1, 24)
    DEP_FIELD(GEM_TSU_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(GEM_TSU_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(GEM_TSU_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(DLL_REF_CTRL, 0x104)
    DEP_FIELD(DLL_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(AMS_REF_CTRL, 0x108)
    DEP_FIELD(AMS_REF_CTRL, CLKACT, 1, 24)
    DEP_FIELD(AMS_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(AMS_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(AMS_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(I2C0_REF_CTRL, 0x120)
    DEP_FIELD(I2C0_REF_CTRL, CLKACT, 1, 24)
    DEP_FIELD(I2C0_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(I2C0_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(I2C0_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(I2C1_REF_CTRL, 0x124)
    DEP_FIELD(I2C1_REF_CTRL, CLKACT, 1, 24)
    DEP_FIELD(I2C1_REF_CTRL, DIVISOR1, 6, 16)
    DEP_FIELD(I2C1_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(I2C1_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(TIMESTAMP_REF_CTRL, 0x128)
    DEP_FIELD(TIMESTAMP_REF_CTRL, CLKACT, 1, 24)
    DEP_FIELD(TIMESTAMP_REF_CTRL, DIVISOR0, 6, 8)
    DEP_FIELD(TIMESTAMP_REF_CTRL, SRCSEL, 3, 0)
DEP_REG32(SAFETY_CHK, 0x130)
    DEP_FIELD(SAFETY_CHK, CHK_VAL, 32, 0)
DEP_REG32(CLKMON_STATUS, 0x140)
    DEP_FIELD(CLKMON_STATUS, CNTA7_OVER_ERR, 1, 15)
    DEP_FIELD(CLKMON_STATUS, MON7_ERR, 1, 14)
    DEP_FIELD(CLKMON_STATUS, CNTA6_OVER_ERR, 1, 13)
    DEP_FIELD(CLKMON_STATUS, MON6_ERR, 1, 12)
    DEP_FIELD(CLKMON_STATUS, CNTA5_OVER_ERR, 1, 11)
    DEP_FIELD(CLKMON_STATUS, MON5_ERR, 1, 10)
    DEP_FIELD(CLKMON_STATUS, CNTA4_OVER_ERR, 1, 9)
    DEP_FIELD(CLKMON_STATUS, MON4_ERR, 1, 8)
    DEP_FIELD(CLKMON_STATUS, CNTA3_OVER_ERR, 1, 7)
    DEP_FIELD(CLKMON_STATUS, MON3_ERR, 1, 6)
    DEP_FIELD(CLKMON_STATUS, CNTA2_OVER_ERR, 1, 5)
    DEP_FIELD(CLKMON_STATUS, MON2_ERR, 1, 4)
    DEP_FIELD(CLKMON_STATUS, CNTA1_OVER_ERR, 1, 3)
    DEP_FIELD(CLKMON_STATUS, MON1_ERR, 1, 2)
    DEP_FIELD(CLKMON_STATUS, CNTA0_OVER_ERR, 1, 1)
    DEP_FIELD(CLKMON_STATUS, MON0_ERR, 1, 0)
DEP_REG32(CLKMON_MASK, 0x144)
    DEP_FIELD(CLKMON_MASK, CNTA7_OVER_ERR, 1, 15)
    DEP_FIELD(CLKMON_MASK, MON7_ERR, 1, 14)
    DEP_FIELD(CLKMON_MASK, CNTA6_OVER_ERR, 1, 13)
    DEP_FIELD(CLKMON_MASK, MON6_ERR, 1, 12)
    DEP_FIELD(CLKMON_MASK, CNTA5_OVER_ERR, 1, 11)
    DEP_FIELD(CLKMON_MASK, MON5_ERR, 1, 10)
    DEP_FIELD(CLKMON_MASK, CNTA4_OVER_ERR, 1, 9)
    DEP_FIELD(CLKMON_MASK, MON4_ERR, 1, 8)
    DEP_FIELD(CLKMON_MASK, CNTA3_OVER_ERR, 1, 7)
    DEP_FIELD(CLKMON_MASK, MON3_ERR, 1, 6)
    DEP_FIELD(CLKMON_MASK, CNTA2_OVER_ERR, 1, 5)
    DEP_FIELD(CLKMON_MASK, MON2_ERR, 1, 4)
    DEP_FIELD(CLKMON_MASK, CNTA1_OVER_ERR, 1, 3)
    DEP_FIELD(CLKMON_MASK, MON1_ERR, 1, 2)
    DEP_FIELD(CLKMON_MASK, CNTA0_OVER_ERR, 1, 1)
    DEP_FIELD(CLKMON_MASK, MON0_ERR, 1, 0)
DEP_REG32(CLKMON_ENABLE, 0x148)
    DEP_FIELD(CLKMON_ENABLE, CNTA7_OVER_ERR, 1, 15)
    DEP_FIELD(CLKMON_ENABLE, MON7_ERR, 1, 14)
    DEP_FIELD(CLKMON_ENABLE, CNTA6_OVER_ERR, 1, 13)
    DEP_FIELD(CLKMON_ENABLE, MON6_ERR, 1, 12)
    DEP_FIELD(CLKMON_ENABLE, CNTA5_OVER_ERR, 1, 11)
    DEP_FIELD(CLKMON_ENABLE, MON5_ERR, 1, 10)
    DEP_FIELD(CLKMON_ENABLE, CNTA4_OVER_ERR, 1, 9)
    DEP_FIELD(CLKMON_ENABLE, MON4_ERR, 1, 8)
    DEP_FIELD(CLKMON_ENABLE, CNTA3_OVER_ERR, 1, 7)
    DEP_FIELD(CLKMON_ENABLE, MON3_ERR, 1, 6)
    DEP_FIELD(CLKMON_ENABLE, CNTA2_OVER_ERR, 1, 5)
    DEP_FIELD(CLKMON_ENABLE, MON2_ERR, 1, 4)
    DEP_FIELD(CLKMON_ENABLE, CNTA1_OVER_ERR, 1, 3)
    DEP_FIELD(CLKMON_ENABLE, MON1_ERR, 1, 2)
    DEP_FIELD(CLKMON_ENABLE, CNTA0_OVER_ERR, 1, 1)
    DEP_FIELD(CLKMON_ENABLE, MON0_ERR, 1, 0)
DEP_REG32(CLKMON_DISABLE, 0x14C)
    DEP_FIELD(CLKMON_DISABLE, CNTA7_OVER_ERR, 1, 15)
    DEP_FIELD(CLKMON_DISABLE, MON7_ERR, 1, 14)
    DEP_FIELD(CLKMON_DISABLE, CNTA6_OVER_ERR, 1, 13)
    DEP_FIELD(CLKMON_DISABLE, MON6_ERR, 1, 12)
    DEP_FIELD(CLKMON_DISABLE, CNTA5_OVER_ERR, 1, 11)
    DEP_FIELD(CLKMON_DISABLE, MON5_ERR, 1, 10)
    DEP_FIELD(CLKMON_DISABLE, CNTA4_OVER_ERR, 1, 9)
    DEP_FIELD(CLKMON_DISABLE, MON4_ERR, 1, 8)
    DEP_FIELD(CLKMON_DISABLE, CNTA3_OVER_ERR, 1, 7)
    DEP_FIELD(CLKMON_DISABLE, MON3_ERR, 1, 6)
    DEP_FIELD(CLKMON_DISABLE, CNTA2_OVER_ERR, 1, 5)
    DEP_FIELD(CLKMON_DISABLE, MON2_ERR, 1, 4)
    DEP_FIELD(CLKMON_DISABLE, CNTA1_OVER_ERR, 1, 3)
    DEP_FIELD(CLKMON_DISABLE, MON1_ERR, 1, 2)
    DEP_FIELD(CLKMON_DISABLE, CNTA0_OVER_ERR, 1, 1)
    DEP_FIELD(CLKMON_DISABLE, MON0_ERR, 1, 0)
DEP_REG32(CLKMON_TRIGGER, 0x150)
    DEP_FIELD(CLKMON_TRIGGER, CNTA7_OVER_ERR, 1, 15)
    DEP_FIELD(CLKMON_TRIGGER, MON7_ERR, 1, 14)
    DEP_FIELD(CLKMON_TRIGGER, CNTA6_OVER_ERR, 1, 13)
    DEP_FIELD(CLKMON_TRIGGER, MON6_ERR, 1, 12)
    DEP_FIELD(CLKMON_TRIGGER, CNTA5_OVER_ERR, 1, 11)
    DEP_FIELD(CLKMON_TRIGGER, MON5_ERR, 1, 10)
    DEP_FIELD(CLKMON_TRIGGER, CNTA4_OVER_ERR, 1, 9)
    DEP_FIELD(CLKMON_TRIGGER, MON4_ERR, 1, 8)
    DEP_FIELD(CLKMON_TRIGGER, CNTA3_OVER_ERR, 1, 7)
    DEP_FIELD(CLKMON_TRIGGER, MON3_ERR, 1, 6)
    DEP_FIELD(CLKMON_TRIGGER, CNTA2_OVER_ERR, 1, 5)
    DEP_FIELD(CLKMON_TRIGGER, MON2_ERR, 1, 4)
    DEP_FIELD(CLKMON_TRIGGER, CNTA1_OVER_ERR, 1, 3)
    DEP_FIELD(CLKMON_TRIGGER, MON1_ERR, 1, 2)
    DEP_FIELD(CLKMON_TRIGGER, CNTA0_OVER_ERR, 1, 1)
    DEP_FIELD(CLKMON_TRIGGER, MON0_ERR, 1, 0)
DEP_REG32(CHKR0_CLKA_UPPER, 0x160)
    DEP_FIELD(CHKR0_CLKA_UPPER, THRSHLD, 32, 0)
DEP_REG32(CHKR0_CLKA_LOWER, 0x164)
    DEP_FIELD(CHKR0_CLKA_LOWER, THRSHLD, 32, 0)
DEP_REG32(CHKR0_CLKB_CNT, 0x168)
    DEP_FIELD(CHKR0_CLKB_CNT, VALUE, 32, 0)
DEP_REG32(CHKR0_CTRL, 0x16C)
    DEP_FIELD(CHKR0_CTRL, START_SINGLE, 1, 8)
    DEP_FIELD(CHKR0_CTRL, START_CONTINUOUS, 1, 7)
    DEP_FIELD(CHKR0_CTRL, CLKB_MUX_CTRL, 1, 5)
    DEP_FIELD(CHKR0_CTRL, CLKA_MUX_CTRL, 3, 1)
    DEP_FIELD(CHKR0_CTRL, ENABLE, 1, 0)
DEP_REG32(CHKR1_CLKA_UPPER, 0x170)
    DEP_FIELD(CHKR1_CLKA_UPPER, THRSHLD, 32, 0)
DEP_REG32(CHKR1_CLKA_LOWER, 0x174)
    DEP_FIELD(CHKR1_CLKA_LOWER, THRSHLD, 32, 0)
DEP_REG32(CHKR1_CLKB_CNT, 0x178)
    DEP_FIELD(CHKR1_CLKB_CNT, VALUE, 32, 0)
DEP_REG32(CHKR1_CTRL, 0x17C)
    DEP_FIELD(CHKR1_CTRL, START_SINGLE, 1, 8)
    DEP_FIELD(CHKR1_CTRL, START_CONTINUOUS, 1, 7)
    DEP_FIELD(CHKR1_CTRL, CLKB_MUX_CTRL, 1, 5)
    DEP_FIELD(CHKR1_CTRL, CLKA_MUX_CTRL, 3, 1)
    DEP_FIELD(CHKR1_CTRL, ENABLE, 1, 0)
DEP_REG32(CHKR2_CLKA_UPPER, 0x180)
    DEP_FIELD(CHKR2_CLKA_UPPER, THRSHLD, 32, 0)
DEP_REG32(CHKR2_CLKA_LOWER, 0x184)
    DEP_FIELD(CHKR2_CLKA_LOWER, THRSHLD, 32, 0)
DEP_REG32(CHKR2_CLKB_CNT, 0x188)
    DEP_FIELD(CHKR2_CLKB_CNT, VALUE, 32, 0)
DEP_REG32(CHKR2_CTRL, 0x18C)
    DEP_FIELD(CHKR2_CTRL, START_SINGLE, 1, 8)
    DEP_FIELD(CHKR2_CTRL, START_CONTINUOUS, 1, 7)
    DEP_FIELD(CHKR2_CTRL, CLKB_MUX_CTRL, 1, 5)
    DEP_FIELD(CHKR2_CTRL, CLKA_MUX_CTRL, 3, 1)
    DEP_FIELD(CHKR2_CTRL, ENABLE, 1, 0)
DEP_REG32(CHKR3_CLKA_UPPER, 0x190)
    DEP_FIELD(CHKR3_CLKA_UPPER, THRSHLD, 32, 0)
DEP_REG32(CHKR3_CLKA_LOWER, 0x194)
    DEP_FIELD(CHKR3_CLKA_LOWER, THRSHLD, 32, 0)
DEP_REG32(CHKR3_CLKB_CNT, 0x198)
    DEP_FIELD(CHKR3_CLKB_CNT, VALUE, 32, 0)
DEP_REG32(CHKR3_CTRL, 0x19C)
    DEP_FIELD(CHKR3_CTRL, START_SINGLE, 1, 8)
    DEP_FIELD(CHKR3_CTRL, START_CONTINUOUS, 1, 7)
    DEP_FIELD(CHKR3_CTRL, CLKB_MUX_CTRL, 1, 5)
    DEP_FIELD(CHKR3_CTRL, CLKA_MUX_CTRL, 3, 1)
    DEP_FIELD(CHKR3_CTRL, ENABLE, 1, 0)
DEP_REG32(CHKR4_CLKA_UPPER, 0x1A0)
    DEP_FIELD(CHKR4_CLKA_UPPER, THRSHLD, 32, 0)
DEP_REG32(CHKR4_CLKA_LOWER, 0x1A4)
    DEP_FIELD(CHKR4_CLKA_LOWER, THRSHLD, 32, 0)
DEP_REG32(CHKR4_CLKB_CNT, 0x1A8)
    DEP_FIELD(CHKR4_CLKB_CNT, VALUE, 32, 0)
DEP_REG32(CHKR4_CTRL, 0x1AC)
    DEP_FIELD(CHKR4_CTRL, START_SINGLE, 1, 8)
    DEP_FIELD(CHKR4_CTRL, START_CONTINUOUS, 1, 7)
    DEP_FIELD(CHKR4_CTRL, CLKB_MUX_CTRL, 1, 5)
    DEP_FIELD(CHKR4_CTRL, CLKA_MUX_CTRL, 3, 1)
    DEP_FIELD(CHKR4_CTRL, ENABLE, 1, 0)
DEP_REG32(CHKR5_CLKA_UPPER, 0x1B0)
    DEP_FIELD(CHKR5_CLKA_UPPER, THRSHLD, 32, 0)
DEP_REG32(CHKR5_CLKA_LOWER, 0x1B4)
    DEP_FIELD(CHKR5_CLKA_LOWER, THRSHLD, 32, 0)
DEP_REG32(CHKR5_CLKB_CNT, 0x1B8)
    DEP_FIELD(CHKR5_CLKB_CNT, VALUE, 32, 0)
DEP_REG32(CHKR5_CTRL, 0x1BC)
    DEP_FIELD(CHKR5_CTRL, START_SINGLE, 1, 8)
    DEP_FIELD(CHKR5_CTRL, START_CONTINUOUS, 1, 7)
    DEP_FIELD(CHKR5_CTRL, CLKB_MUX_CTRL, 1, 5)
    DEP_FIELD(CHKR5_CTRL, CLKA_MUX_CTRL, 3, 1)
    DEP_FIELD(CHKR5_CTRL, ENABLE, 1, 0)
DEP_REG32(CHKR6_CLKA_UPPER, 0x1C0)
    DEP_FIELD(CHKR6_CLKA_UPPER, THRSHLD, 32, 0)
DEP_REG32(CHKR6_CLKA_LOWER, 0x1C4)
    DEP_FIELD(CHKR6_CLKA_LOWER, THRSHLD, 32, 0)
DEP_REG32(CHKR6_CLKB_CNT, 0x1C8)
    DEP_FIELD(CHKR6_CLKB_CNT, VALUE, 32, 0)
DEP_REG32(CHKR6_CTRL, 0x1CC)
    DEP_FIELD(CHKR6_CTRL, START_SINGLE, 1, 8)
    DEP_FIELD(CHKR6_CTRL, START_CONTINUOUS, 1, 7)
    DEP_FIELD(CHKR6_CTRL, CLKB_MUX_CTRL, 1, 5)
    DEP_FIELD(CHKR6_CTRL, CLKA_MUX_CTRL, 3, 1)
    DEP_FIELD(CHKR6_CTRL, ENABLE, 1, 0)
DEP_REG32(CHKR7_CLKA_UPPER, 0x1D0)
    DEP_FIELD(CHKR7_CLKA_UPPER, THRSHLD, 32, 0)
DEP_REG32(CHKR7_CLKA_LOWER, 0x1D4)
    DEP_FIELD(CHKR7_CLKA_LOWER, THRSHLD, 32, 0)
DEP_REG32(CHKR7_CLKB_CNT, 0x1D8)
    DEP_FIELD(CHKR7_CLKB_CNT, VALUE, 32, 0)
DEP_REG32(CHKR7_CTRL, 0x1DC)
    DEP_FIELD(CHKR7_CTRL, START_SINGLE, 1, 8)
    DEP_FIELD(CHKR7_CTRL, START_CONTINUOUS, 1, 7)
    DEP_FIELD(CHKR7_CTRL, CLKB_MUX_CTRL, 1, 5)
    DEP_FIELD(CHKR7_CTRL, CLKA_MUX_CTRL, 3, 1)
    DEP_FIELD(CHKR7_CTRL, ENABLE, 1, 0)
DEP_REG32(BOOT_MODE_USER, 0x200)
    DEP_FIELD(BOOT_MODE_USER, ALT_BOOT_MODE, 4, 12)
    DEP_FIELD(BOOT_MODE_USER, USE_ALT, 1, 8)
    DEP_FIELD(BOOT_MODE_USER, BOOT_MODE, 4, 0)
DEP_REG32(BOOT_MODE_POR, 0x204)
    DEP_FIELD(BOOT_MODE_POR, BOOT_MODE2, 4, 8)
    DEP_FIELD(BOOT_MODE_POR, BOOT_MODE1, 4, 4)
    DEP_FIELD(BOOT_MODE_POR, BOOT_MODE0, 4, 0)
DEP_REG32(RESET_CTRL, 0x218)
    DEP_FIELD(RESET_CTRL, SOFT_RESET, 1, 4)
    DEP_FIELD(RESET_CTRL, SRST_DIS, 1, 0)
DEP_REG32(BLOCKONLY_RST, 0x21c)
    DEP_FIELD(BLOCKONLY_RST, DEBUG_ONLY, 1, 0)
DEP_REG32(RESET_REASON, 0x220)
    DEP_FIELD(RESET_REASON, DEBUG_SYS, 1, 6)
    DEP_FIELD(RESET_REASON, SOFT, 1, 5)
    DEP_FIELD(RESET_REASON, SRST, 1, 4)
    DEP_FIELD(RESET_REASON, PSONLY_RESET_REQ, 1, 3)
    DEP_FIELD(RESET_REASON, PMU_SYS_RESET, 1, 2)
    DEP_FIELD(RESET_REASON, EXTERNAL_POR, 1, 0)
DEP_REG32(RST_LPD_IOU0, 0x230)
    DEP_FIELD(RST_LPD_IOU0, GEM3_RESET, 1, 3)
    DEP_FIELD(RST_LPD_IOU0, GEM2_RESET, 1, 2)
    DEP_FIELD(RST_LPD_IOU0, GEM1_RESET, 1, 1)
    DEP_FIELD(RST_LPD_IOU0, GEM0_RESET, 1, 0)
DEP_REG32(RST_LPD_IOU1, 0x234)
DEP_REG32(RST_LPD_IOU2, 0x238)
    DEP_FIELD(RST_LPD_IOU2, TIMESTAMP_RESET, 1, 20)
    DEP_FIELD(RST_LPD_IOU2, IOU_CC_RESET, 1, 19)
    DEP_FIELD(RST_LPD_IOU2, GPIO_RESET, 1, 18)
    DEP_FIELD(RST_LPD_IOU2, ADMA_RESET, 1, 17)
    DEP_FIELD(RST_LPD_IOU2, NAND_RESET, 1, 16)
    DEP_FIELD(RST_LPD_IOU2, SWDT_RESET, 1, 15)
    DEP_FIELD(RST_LPD_IOU2, TTC3_RESET, 1, 14)
    DEP_FIELD(RST_LPD_IOU2, TTC2_RESET, 1, 13)
    DEP_FIELD(RST_LPD_IOU2, TTC1_RESET, 1, 12)
    DEP_FIELD(RST_LPD_IOU2, TTC0_RESET, 1, 11)
    DEP_FIELD(RST_LPD_IOU2, I2C1_RESET, 1, 10)
    DEP_FIELD(RST_LPD_IOU2, I2C0_RESET, 1, 9)
    DEP_FIELD(RST_LPD_IOU2, CAN1_RESET, 1, 8)
    DEP_FIELD(RST_LPD_IOU2, CAN0_RESET, 1, 7)
    DEP_FIELD(RST_LPD_IOU2, SDIO1_RESET, 1, 6)
    DEP_FIELD(RST_LPD_IOU2, SDIO0_RESET, 1, 5)
    DEP_FIELD(RST_LPD_IOU2, SPI1_RESET, 1, 4)
    DEP_FIELD(RST_LPD_IOU2, SPI0_RESET, 1, 3)
    DEP_FIELD(RST_LPD_IOU2, UART1_RESET, 1, 2)
    DEP_FIELD(RST_LPD_IOU2, UART0_RESET, 1, 1)
    DEP_FIELD(RST_LPD_IOU2, QSPI_RESET, 1, 0)
DEP_REG32(RST_LPD_TOP, 0x23c)
    DEP_FIELD(RST_LPD_TOP, FPD_RESET, 1, 23)
    DEP_FIELD(RST_LPD_TOP, LPD_SWDT_RESET, 1, 20)
    DEP_FIELD(RST_LPD_TOP, AFI_FM6_RESET, 1, 19)
    DEP_FIELD(RST_LPD_TOP, SYSMON_RESET, 1, 17)
    DEP_FIELD(RST_LPD_TOP, RTC_RESET, 1, 16)
    DEP_FIELD(RST_LPD_TOP, APM_RESET, 1, 15)
    DEP_FIELD(RST_LPD_TOP, IPI_RESET, 1, 14)
    DEP_FIELD(RST_LPD_TOP, USB1_APB_RESET, 1, 11)
    DEP_FIELD(RST_LPD_TOP, USB0_APB_RESET, 1, 10)
    DEP_FIELD(RST_LPD_TOP, USB1_HIBERRESET, 1, 9)
    DEP_FIELD(RST_LPD_TOP, USB0_HIBERRESET, 1, 8)
    DEP_FIELD(RST_LPD_TOP, USB1_CORERESET, 1, 7)
    DEP_FIELD(RST_LPD_TOP, USB0_CORERESET, 1, 6)
    DEP_FIELD(RST_LPD_TOP, RPU_PGE_RESET, 1, 4)
    DEP_FIELD(RST_LPD_TOP, RPU_OCM_RESET, 1, 3)
    DEP_FIELD(RST_LPD_TOP, RPU_AMBA_RESET, 1, 2)
    DEP_FIELD(RST_LPD_TOP, RPU_R51_RESET, 1, 1)
    DEP_FIELD(RST_LPD_TOP, RPU_R50_RESET, 1, 0)
DEP_REG32(RST_LPD_DBG, 0x240)
    DEP_FIELD(RST_LPD_DBG, DBG_ACK, 1, 15)
    DEP_FIELD(RST_LPD_DBG, RPU_DBG1_RESET, 1, 5)
    DEP_FIELD(RST_LPD_DBG, RPU_DBG0_RESET, 1, 4)
    DEP_FIELD(RST_LPD_DBG, DBG_LPD_RESET, 1, 1)
    DEP_FIELD(RST_LPD_DBG, DBG_FPD_RESET, 1, 0)
DEP_REG32(BANK3_CTRL0, 0x270)
    DEP_FIELD(BANK3_CTRL0, DRIVE0, 10, 0)
DEP_REG32(BANK3_CTRL1, 0x274)
    DEP_FIELD(BANK3_CTRL1, DRIVE1, 10, 0)
DEP_REG32(BANK3_CTRL2, 0x278)
    DEP_FIELD(BANK3_CTRL2, SCHMITT_CMOS_N, 10, 0)
DEP_REG32(BANK3_CTRL3, 0x27C)
    DEP_FIELD(BANK3_CTRL3, PULL_HIGH_LOW_N, 10, 0)
DEP_REG32(BANK3_CTRL4, 0x280)
    DEP_FIELD(BANK3_CTRL4, PULL_ENABLE, 10, 0)
DEP_REG32(BANK3_CTRL5, 0x284)
    DEP_FIELD(BANK3_CTRL5, SLOW_FAST_SLEW_N, 10, 0)
DEP_REG32(BANK3_STATUS, 0x288)
    DEP_FIELD(BANK3_STATUS, VMODE_1P8_3P3_N, 1, 0)

#define R_MAX (R_BANK3_STATUS + 1)

typedef struct CRL_APB {
    SysBusDevice parent_obj;
    MemoryRegion iomem;
    qemu_irq irq_ir;
    qemu_irq mon_irq_ir;

    uint32_t regs[R_MAX];
    DepRegisterInfo regs_info[R_MAX];
} CRL_APB;

static void ir_update_irq(CRL_APB *s)
{
    bool pending = s->regs[R_IR_STATUS] & ~s->regs[R_IR_MASK];
    qemu_set_irq(s->irq_ir, pending);
}

static void ir_status_postw(DepRegisterInfo *reg, uint64_t val64)
{
    CRL_APB *s = XILINX_CRL_APB(reg->opaque);
    ir_update_irq(s);
}

static uint64_t ir_enable_prew(DepRegisterInfo *reg, uint64_t val64)
{
    CRL_APB *s = XILINX_CRL_APB(reg->opaque);
    uint32_t val = val64;

    s->regs[R_IR_MASK] &= ~val;
    ir_update_irq(s);
    return 0;
}

static uint64_t ir_disable_prew(DepRegisterInfo *reg, uint64_t val64)
{
    CRL_APB *s = XILINX_CRL_APB(reg->opaque);
    uint32_t val = val64;

    s->regs[R_IR_MASK] |= val;
    ir_update_irq(s);
    return 0;
}

static void mon_ir_update_irq(CRL_APB *s)
{
    bool pending = s->regs[R_CLKMON_STATUS] & ~s->regs[R_CLKMON_MASK];
    qemu_set_irq(s->mon_irq_ir, pending);
}

static void mon_ir_status_postw(DepRegisterInfo *reg, uint64_t val64)
{
    CRL_APB *s = XILINX_CRL_APB(reg->opaque);

    mon_ir_update_irq(s);
}

static uint64_t mon_ir_enable_prew(DepRegisterInfo *reg, uint64_t val64)
{
    CRL_APB *s = XILINX_CRL_APB(reg->opaque);
    uint32_t val = val64;

    s->regs[R_CLKMON_MASK] &= ~val;
    mon_ir_update_irq(s);
    return 0;
}

static uint64_t mon_ir_disable_prew(DepRegisterInfo *reg, uint64_t val64)
{
    CRL_APB *s = XILINX_CRL_APB(reg->opaque);
    uint32_t val = val64;

    s->regs[R_CLKMON_MASK] |= val;
    mon_ir_update_irq(s);
    return 0;
}

static uint64_t mon_ir_trigger_prew(DepRegisterInfo *reg, uint64_t val64)
{
    CRL_APB *s = XILINX_CRL_APB(reg->opaque);
    uint32_t val = val64;

    s->regs[R_CLKMON_STATUS] = s->regs[R_CLKMON_STATUS] | val;
    mon_ir_update_irq(s);
    return 0;
}

static void update_reset_reason_postw(DepRegisterInfo *reg, uint64_t val64)
{
    CRL_APB *s = XILINX_CRL_APB(reg->opaque);

    DEP_AF_DP32(s->regs, RESET_REASON, SOFT,
                DEP_AF_EX32(s->regs, RESET_CTRL, SOFT_RESET));
    DEP_AF_DP32(s->regs, RESET_REASON, DEBUG_SYS,
                DEP_AF_EX32(s->regs, BLOCKONLY_RST, DEBUG_ONLY));
}

static DepRegisterAccessInfo crl_apb_regs_info[] = {
    {   .name = "ERR_CTRL",  .decode.addr = A_ERR_CTRL,
    },{ .name = "IR_STATUS",  .decode.addr = A_IR_STATUS,
        .w1c = 0x1,
        .post_write = ir_status_postw,
    },{ .name = "IR_MASK",  .decode.addr = A_IR_MASK,
        .reset = 0x1,
        .ro = 0x1,
    },{ .name = "IR_ENABLE",  .decode.addr = A_IR_ENABLE,
        .pre_write = ir_enable_prew,
    },{ .name = "IR_DISABLE",  .decode.addr = A_IR_DISABLE,
        .pre_write = ir_disable_prew,
    },{ .name = "CRL_WPROT",  .decode.addr = A_CRL_WPROT,
    },{ .name = "IOPLL_CTRL",  .decode.addr = A_IOPLL_CTRL,
        .reset = 0x2809,
        .rsvd = 0xf88c80f6L,
    },{ .name = "IOPLL_CFG",  .decode.addr = A_IOPLL_CFG,
        .rsvd = 0x1801210,
    },{ .name = "IOPLL_FRAC_CFG",  .decode.addr = A_IOPLL_FRAC_CFG,
        .rsvd = 0x7e330000,
    },{ .name = "RPLL_CTRL",  .decode.addr = A_RPLL_CTRL,
        .reset = 0x2809,
        .rsvd = 0xf88c80f6L,
    },{ .name = "RPLL_CFG",  .decode.addr = A_RPLL_CFG,
        .rsvd = 0x1801210,
    },{ .name = "RPLL_FRAC_CFG",  .decode.addr = A_RPLL_FRAC_CFG,
        .rsvd = 0x7e330000,
    },{ .name = "PLL_STATUS",  .decode.addr = A_PLL_STATUS,
        .reset = 0x1b,
        .rsvd = 0xffffffc4L,
        .ro = 0x1b,
    },{ .name = "IOPLL_TO_FPD_CTRL",  .decode.addr = A_IOPLL_TO_FPD_CTRL,
        .reset = 0x400,
        .rsvd = 0xc0ff,
    },{ .name = "RPLL_TO_FPD_CTRL",  .decode.addr = A_RPLL_TO_FPD_CTRL,
        .reset = 0x400,
        .rsvd = 0xc0ff,
    },{ .name = "USB3_DUAL_REF_CTRL",  .decode.addr = A_USB3_DUAL_REF_CTRL,
        .reset = 0x52000,
        .rsvd = 0xfdc0c0f8L,
    },{ .name = "GEM0_REF_CTRL",  .decode.addr = A_GEM0_REF_CTRL,
        .reset = 0x2500,
        .rsvd = 0xf9c0c0f8L,
    },{ .name = "GEM1_REF_CTRL",  .decode.addr = A_GEM1_REF_CTRL,
        .reset = 0x2500,
        .rsvd = 0xf9c0c0f8L,
    },{ .name = "GEM2_REF_CTRL",  .decode.addr = A_GEM2_REF_CTRL,
        .reset = 0x2500,
        .rsvd = 0xf9c0c0f8L,
    },{ .name = "GEM3_REF_CTRL",  .decode.addr = A_GEM3_REF_CTRL,
        .reset = 0x2500,
        .rsvd = 0xf9c0c0f8L,
    },{ .name = "USB0_BUS_REF_CTRL",  .decode.addr = A_USB0_BUS_REF_CTRL,
        .reset = 0x52000,
        .rsvd = 0xfdc0c0f8L,
    },{ .name = "USB1_BUS_REF_CTRL",  .decode.addr = A_USB1_BUS_REF_CTRL,
        .reset = 0x12200,
        .rsvd = 0xfdc0c0f8L,
    },{ .name = "QSPI_REF_CTRL",  .decode.addr = A_QSPI_REF_CTRL,
        .reset = 0x1000800,
        .rsvd = 0xfec0c0f8L,
    },{ .name = "SDIO0_REF_CTRL",  .decode.addr = A_SDIO0_REF_CTRL,
        .reset = 0x1003000,
        .rsvd = 0xfeffc0f8L,
    },{ .name = "SDIO1_REF_CTRL",  .decode.addr = A_SDIO1_REF_CTRL,
        .reset = 0x1003f00,
        .rsvd = 0xfeffc0f8L,
    },{ .name = "UART0_REF_CTRL",  .decode.addr = A_UART0_REF_CTRL,
        .reset = 0x1000500,
        .rsvd = 0xfeffc0f8L,
    },{ .name = "UART1_REF_CTRL",  .decode.addr = A_UART1_REF_CTRL,
        .reset = 0x1500,
        .rsvd = 0xfeffc0f8L,
    },{ .name = "SPI0_REF_CTRL",  .decode.addr = A_SPI0_REF_CTRL,
        .reset = 0x1000500,
        .rsvd = 0xfeffc0f8L,
    },{ .name = "SPI1_REF_CTRL",  .decode.addr = A_SPI1_REF_CTRL,
        .reset = 0x1500,
        .rsvd = 0xfeffc0f8L,
    },{ .name = "CAN0_REF_CTRL",  .decode.addr = A_CAN0_REF_CTRL,
        .reset = 0x1000500,
        .rsvd = 0xfeffc0f8L,
    },{ .name = "CAN1_REF_CTRL",  .decode.addr = A_CAN1_REF_CTRL,
        .reset = 0x1500,
        .rsvd = 0xfeffc0f8L,
    },{ .name = "CPU_R5_CTRL",  .decode.addr = A_CPU_R5_CTRL,
        .reset = 0x600,
        .rsvd = 0xfeffc0f8L,
    },{ .name = "IOU_SWITCH_CTRL",  .decode.addr = A_IOU_SWITCH_CTRL,
        .reset = 0x1500,
        .rsvd = 0xfeffc0f8L,
    },{ .name = "CSU_PLL_CTRL",  .decode.addr = A_CSU_PLL_CTRL,
        .reset = 0x1001504,
        .rsvd = 0xfeffc0f8L,
    },{ .name = "PCAP_CTRL",  .decode.addr = A_PCAP_CTRL,
        .reset = 0x1500,
        .rsvd = 0xfeffc0f8L,
    },{ .name = "LPD_SWITCH_CTRL",  .decode.addr = A_LPD_SWITCH_CTRL,
        .reset = 0x1002000,
        .rsvd = 0xfeffc0f8L,
    },{ .name = "LPD_LSBUS_CTRL",  .decode.addr = A_LPD_LSBUS_CTRL,
        .reset = 0x1002000,
        .rsvd = 0xfeffc0f8L,
    },{ .name = "DBG_LPD_CTRL",  .decode.addr = A_DBG_LPD_CTRL,
        .reset = 0x2000,
        .rsvd = 0xfeffc0f8L,
    },{ .name = "NAND_REF_CTRL",  .decode.addr = A_NAND_REF_CTRL,
        .reset = 0x52000,
        .rsvd = 0xfec0c0f8L,
    },{ .name = "ADMA_REF_CTRL",  .decode.addr = A_ADMA_REF_CTRL,
        .reset = 0x2000,
        .rsvd = 0xfeffc0f8L,
    },{ .name = "PL0_REF_CTRL",  .decode.addr = A_PL0_REF_CTRL,
        .reset = 0x52000,
        .rsvd = 0xfec0c0f8L,
    },{ .name = "PL1_REF_CTRL",  .decode.addr = A_PL1_REF_CTRL,
        .reset = 0x52000,
        .rsvd = 0xfec0c0f8L,
    },{ .name = "PL2_REF_CTRL",  .decode.addr = A_PL2_REF_CTRL,
        .reset = 0x52000,
        .rsvd = 0xfec0c0f8L,
    },{ .name = "PL3_REF_CTRL",  .decode.addr = A_PL3_REF_CTRL,
        .reset = 0x52000,
        .rsvd = 0xfec0c0f8L,
    },{ .name = "PL0_THR_CTRL",  .decode.addr = A_PL0_THR_CTRL,
        .reset = 0x1,
        .rsvd = 0x7ffc,
        .ro = 0xffff8000L,
    },{ .name = "PL0_THR_CNT",  .decode.addr = A_PL0_THR_CNT,
    },{ .name = "PL1_THR_CTRL",  .decode.addr = A_PL1_THR_CTRL,
        .reset = 0x1,
        .rsvd = 0x7ffc,
        .ro = 0xffff8000L,
    },{ .name = "PL1_THR_CNT",  .decode.addr = A_PL1_THR_CNT,
    },{ .name = "PL2_THR_CTRL",  .decode.addr = A_PL2_THR_CTRL,
        .reset = 0x1,
        .rsvd = 0x7ffc,
        .ro = 0xffff8000L,
    },{ .name = "PL2_THR_CNT",  .decode.addr = A_PL2_THR_CNT,
    },{ .name = "PL3_THR_CTRL",  .decode.addr = A_PL3_THR_CTRL,
        .reset = 0x1,
        .rsvd = 0x7ffc,
        .ro = 0xffff8000L,
    },{ .name = "PL3_THR_CNT",  .decode.addr = A_PL3_THR_CNT,
    },{ .name = "GEM_TSU_REF_CTRL",  .decode.addr = A_GEM_TSU_REF_CTRL,
        .reset = 0x51000,
        .rsvd = 0xfec0c0f8L,
    },{ .name = "DLL_REF_CTRL",  .decode.addr = A_DLL_REF_CTRL,
        .rsvd = 0xf8,
    },{ .name = "AMS_REF_CTRL",  .decode.addr = A_AMS_REF_CTRL,
        .reset = 0x1800,
        .rsvd = 0xfeffc0f8L,
    },{ .name = "I2C0_REF_CTRL",  .decode.addr = A_I2C0_REF_CTRL,
        .reset = 0x01000500,
        .rsvd = 0xfec0c0f8L,
    },{ .name = "I2C1_REF_CTRL",  .decode.addr = A_I2C1_REF_CTRL,
        .reset = 0x01000500,
        .rsvd = 0xfec0c0f8L,
    },{ .name = "TIMESTAMP_REF_CTRL",  .decode.addr = A_TIMESTAMP_REF_CTRL,
        .reset = 0x01001800,
        .rsvd = 0xfeffc0f8L,
    },{ .name = "SAFETY_CHK",  .decode.addr = A_SAFETY_CHK,
        .reset = 0x00000000,
    },{ .name = "CLKMON_STATUS",  .decode.addr = A_CLKMON_STATUS,
        .reset = 0x00000000,
        .w1c = 0xFFFF,
        .post_write = mon_ir_status_postw,
    },{ .name = "CLKMON_MASK",  .decode.addr = A_CLKMON_MASK,
        .reset = 0x0000FFFF,
        .ro = 0xFFFF,
    },{ .name = "CLKMON_ENABLE",  .decode.addr = A_CLKMON_ENABLE,
        .pre_write = mon_ir_enable_prew,
    },{ .name = "CLKMON_DISABLE",  .decode.addr = A_CLKMON_DISABLE,
        .pre_write = mon_ir_disable_prew,
    },{ .name = "CLKMON_TRIGGER",  .decode.addr = A_CLKMON_TRIGGER,
        .pre_write = mon_ir_trigger_prew,
    },{ .name = "CHKR0_CLKA_UPPER",  .decode.addr = A_CHKR0_CLKA_UPPER,
    },{ .name = "CHKR0_CLKA_LOWER",  .decode.addr = A_CHKR0_CLKA_LOWER,
    },{ .name = "CHKR0_CLKB_CNT",  .decode.addr = A_CHKR0_CLKB_CNT,
    },{ .name = "CHKR0_CTRL",  .decode.addr = A_CHKR0_CTRL,
    },{ .name = "CHKR1_CLKA_UPPER",  .decode.addr = A_CHKR1_CLKA_UPPER,
    },{ .name = "CHKR1_CLKA_LOWER",  .decode.addr = A_CHKR1_CLKA_LOWER,
    },{ .name = "CHKR1_CLKB_CNT",  .decode.addr = A_CHKR1_CLKB_CNT,
    },{ .name = "CHKR1_CTRL",  .decode.addr = A_CHKR1_CTRL,
    },{ .name = "CHKR2_CLKA_UPPER",  .decode.addr = A_CHKR2_CLKA_UPPER,
    },{ .name = "CHKR2_CLKA_LOWER",  .decode.addr = A_CHKR2_CLKA_LOWER,
    },{ .name = "CHKR2_CLKB_CNT",  .decode.addr = A_CHKR2_CLKB_CNT,
    },{ .name = "CHKR2_CTRL",  .decode.addr = A_CHKR2_CTRL,
    },{ .name = "CHKR3_CLKA_UPPER",  .decode.addr = A_CHKR3_CLKA_UPPER,
    },{ .name = "CHKR3_CLKA_LOWER",  .decode.addr = A_CHKR3_CLKA_LOWER,
    },{ .name = "CHKR3_CLKB_CNT",  .decode.addr = A_CHKR3_CLKB_CNT,
    },{ .name = "CHKR3_CTRL",  .decode.addr = A_CHKR3_CTRL,
    },{ .name = "CHKR4_CLKA_UPPER",  .decode.addr = A_CHKR4_CLKA_UPPER,
    },{ .name = "CHKR4_CLKA_LOWER",  .decode.addr = A_CHKR4_CLKA_LOWER,
    },{ .name = "CHKR4_CLKB_CNT",  .decode.addr = A_CHKR4_CLKB_CNT,
    },{ .name = "CHKR4_CTRL",  .decode.addr = A_CHKR4_CTRL,
    },{ .name = "CHKR5_CLKA_UPPER",  .decode.addr = A_CHKR5_CLKA_UPPER,
    },{ .name = "CHKR5_CLKA_LOWER",  .decode.addr = A_CHKR5_CLKA_LOWER,
    },{ .name = "CHKR5_CLKB_CNT",  .decode.addr = A_CHKR5_CLKB_CNT,
    },{ .name = "CHKR5_CTRL",  .decode.addr = A_CHKR5_CTRL,
    },{ .name = "CHKR6_CLKA_UPPER",  .decode.addr = A_CHKR6_CLKA_UPPER,
    },{ .name = "CHKR6_CLKA_LOWER",  .decode.addr = A_CHKR6_CLKA_LOWER,
    },{ .name = "CHKR6_CLKB_CNT",  .decode.addr = A_CHKR6_CLKB_CNT,
    },{ .name = "CHKR6_CTRL",  .decode.addr = A_CHKR6_CTRL,
    },{ .name = "CHKR7_CLKA_UPPER",  .decode.addr = A_CHKR7_CLKA_UPPER,
    },{ .name = "CHKR7_CLKA_LOWER",  .decode.addr = A_CHKR7_CLKA_LOWER,
    },{ .name = "CHKR7_CLKB_CNT",  .decode.addr = A_CHKR7_CLKB_CNT,
    },{ .name = "CHKR7_CTRL",  .decode.addr = A_CHKR7_CTRL,
    },
    {   .name = "BOOT_MODE_USER",  .decode.addr = A_BOOT_MODE_USER,
        .reset = 0x4,
        .rsvd = 0xf0,
    },{ .name = "BOOT_MODE_POR",  .decode.addr = A_BOOT_MODE_POR,
        .rsvd = 0xf000,
        .ro = 0xfff,
    },{ .name = "RESET_CTRL",  .decode.addr = A_RESET_CTRL,
        .reset = 0x40,
        .rsvd = 0xff8c,
        .gpios = (DepRegisterGPIOMapping[]) {
            { .name = "SRST_B", .bit_pos = 4,   .width = 1 },
            {},
        },
        .post_write = update_reset_reason_postw,
    },{ .name = "BLOCKONLY_RST",  .decode.addr = A_BLOCKONLY_RST,
        .rsvd = 0x7dcc,
        .ro = 0x7dcc,
        .w1c = 0x8233,
        .post_write = update_reset_reason_postw,
    },{ .name = "RESET_REASON",  .decode.addr = A_RESET_REASON,
        .rsvd = 0x7e0c,
        .ro = 0x7e0c,
        .w1c = 0x81f3,
        .reset = 0x1,
    },{ .name = "RST_LPD_IOU0",  .decode.addr = A_RST_LPD_IOU0,
        .reset = 0xf,
        .rsvd = 0xfff0,
    },{ .name = "RST_LPD_IOU1",  .decode.addr = A_RST_LPD_IOU1,
        .rsvd = 0xff,
    },{ .name = "RST_LPD_IOU2",  .decode.addr = A_RST_LPD_IOU2,
        .reset = 0x7ffff,
        .rsvd = 0xfff00000L,
    },{ .name = "RST_LPD_TOP",  .decode.addr = A_RST_LPD_TOP,
        .reset = 0x188fdf,
        .rsvd = 0x643020,
        .gpios = (DepRegisterGPIOMapping[]) {
            { .name = "RST_R5", .bit_pos = 0,   .num = 2 },
            {},
        },
        .inhibit_reset = 1u << 31,
    },{ .name = "RST_LPD_DBG",  .decode.addr = A_RST_LPD_DBG,
        .reset = 0x33,
        .rsvd = 0xffcc,
    },{ .name = "BANK3_CTRL0",  .decode.addr = A_BANK3_CTRL0,
        .reset = 0x3ff,
    },{ .name = "BANK3_CTRL1",  .decode.addr = A_BANK3_CTRL1,
        .reset = 0x3ff,
    },{ .name = "BANK3_CTRL2",  .decode.addr = A_BANK3_CTRL2,
        .reset = 0x3ff,
    },{ .name = "BANK3_CTRL3",  .decode.addr = A_BANK3_CTRL3,
        .reset = 0x3ff,
    },{ .name = "BANK3_CTRL4",  .decode.addr = A_BANK3_CTRL4,
        .reset = 0x3ff,
    },{ .name = "BANK3_CTRL5",  .decode.addr = A_BANK3_CTRL5,
        .reset = 0x3ff,
    },{ .name = "BANK3_STATUS",  .decode.addr = A_BANK3_STATUS,
        .reset = 0,
        .ro = 1
    }
};

static void crl_apb_reset(DeviceState *dev)
{
    CRL_APB *s = XILINX_CRL_APB(dev);
    unsigned int i;
    QemuOpts *opts = qemu_find_opts_singleton("boot-opts");
    uint32_t boot_mode = qemu_opt_get_number(opts, "mode", 0);
    uint32_t reset_reason_value = s->regs[R_RESET_REASON];

    assert(boot_mode < (1 << R_BOOT_MODE_USER_BOOT_MODE_LENGTH));

    for (i = 0; i < ARRAY_SIZE(s->regs_info); ++i) {
        dep_register_reset(&s->regs_info[i]);
    }

    s->regs[R_RESET_REASON] = reset_reason_value;

    s->regs[R_BOOT_MODE_POR] = deposit32(s->regs[R_BOOT_MODE_POR],
                                         R_BOOT_MODE_POR_BOOT_MODE0_SHIFT,
                                         R_BOOT_MODE_POR_BOOT_MODE0_LENGTH,
                                         boot_mode);

    s->regs[R_BOOT_MODE_USER] =     deposit32(s->regs[R_BOOT_MODE_USER],
                                         R_BOOT_MODE_USER_BOOT_MODE_SHIFT,
                                         R_BOOT_MODE_USER_BOOT_MODE_LENGTH,
                                         boot_mode);
    ir_update_irq(s);
}

static uint64_t crl_apb_read(void *opaque, hwaddr addr, unsigned size)
{
    CRL_APB *s = XILINX_CRL_APB(opaque);
    DepRegisterInfo *r = &s->regs_info[addr / 4];

    if (!r->data) {
        qemu_log_mask(LOG_GUEST_ERROR,
                 "%s: Decode error: read from 0x%" HWADDR_PRIx "\n",
                 object_get_canonical_path(OBJECT(s)),
                 addr);
        return 0;
    }
    return dep_register_read(r);
}

static void crl_apb_write(void *opaque, hwaddr addr, uint64_t value,
                      unsigned size)
{
    CRL_APB *s = XILINX_CRL_APB(opaque);
    DepRegisterInfo *r = &s->regs_info[addr / 4];

    if (!r->data) {
        qemu_log_mask(LOG_GUEST_ERROR,
                 "%s: Decode error: write from 0x%" HWADDR_PRIx "=0x%" PRIx64 "\n",
                 object_get_canonical_path(OBJECT(s)),
                 addr, value);
        return;
    }
    dep_register_write(r, value, ~0);
}

static const MemoryRegionOps crl_apb_ops = {
    .read = crl_apb_read,
    .write = crl_apb_write,
    .endianness = DEVICE_LITTLE_ENDIAN,
    .valid = {
        .min_access_size = 4,
        .max_access_size = 4,
    },
};

static void crl_apb_realize(DeviceState *dev, Error **errp)
{
    CRL_APB *s = XILINX_CRL_APB(dev);
    const char *prefix = object_get_canonical_path(OBJECT(dev));
    unsigned int i;

    for (i = 0; i < ARRAY_SIZE(crl_apb_regs_info); ++i) {
        DepRegisterInfo *r = &s->regs_info[crl_apb_regs_info[i].decode.addr/4];

        *r = (DepRegisterInfo) {
            .data = (uint8_t *)&s->regs[
                    crl_apb_regs_info[i].decode.addr/4],
            .data_size = sizeof(uint32_t),
            .access = &crl_apb_regs_info[i],
            .debug = XILINX_CRL_APB_ERR_DEBUG,
            .prefix = prefix,
            .opaque = s,
        };
        dep_register_init(r);
        qdev_pass_all_gpios(DEVICE(r), dev);
    }
}

static void crl_apb_init(Object *obj)
{
    CRL_APB *s = XILINX_CRL_APB(obj);
    SysBusDevice *sbd = SYS_BUS_DEVICE(obj);

    memory_region_init_io(&s->iomem, obj, &crl_apb_ops, s,
                          TYPE_XILINX_CRL_APB, R_MAX * 4);
    sysbus_init_mmio(sbd, &s->iomem);
    sysbus_init_irq(sbd, &s->irq_ir);
    qdev_init_gpio_out_named(DEVICE(obj), &s->mon_irq_ir, "clkmon_error_out",
                             1);
}

static const VMStateDescription vmstate_crl_apb = {
    .name = TYPE_XILINX_CRL_APB,
    .version_id = 1,
    .minimum_version_id = 1,
    .minimum_version_id_old = 1,
    .fields = (VMStateField[]) {
        VMSTATE_UINT32_ARRAY(regs, CRL_APB, R_MAX),
        VMSTATE_END_OF_LIST(),
    }
};

static const FDTGenericGPIOSet crl_gpios[] = {
    {
        .names = &fdt_generic_gpio_name_set_gpio,
        .gpios = (FDTGenericGPIOConnection[]) {
            { .name = "RST_R5",     .fdt_index = 0,     .range = 2 },
            { .name = "SRST_B",     .fdt_index = 2  },
            { },
        }
    },
    { },
};

static const FDTGenericGPIOSet crl_client_gpios [] = {
    {
        .names = &fdt_generic_gpio_name_set_gpio,
        .gpios = (FDTGenericGPIOConnection [])  {
            { .name = "clkmon_error_out", .fdt_index = 0, .range = 1 },
            { },
        },
    },
    { },
};

static void crl_apb_class_init(ObjectClass *klass, void *data)
{
    DeviceClass *dc = DEVICE_CLASS(klass);
    FDTGenericGPIOClass *fggc = FDT_GENERIC_GPIO_CLASS(klass);

    dc->reset = crl_apb_reset;
    dc->realize = crl_apb_realize;
    dc->vmsd = &vmstate_crl_apb;
    fggc->controller_gpios = crl_gpios;
    fggc->client_gpios = crl_client_gpios;
}

static const TypeInfo crl_apb_info = {
    .name          = TYPE_XILINX_CRL_APB,
    .parent        = TYPE_SYS_BUS_DEVICE,
    .instance_size = sizeof(CRL_APB),
    .class_init    = crl_apb_class_init,
    .instance_init = crl_apb_init,
    .interfaces    = (InterfaceInfo[]) {
        { TYPE_FDT_GENERIC_GPIO },
        { }
    },
};

static void crl_apb_register_types(void)
{
    type_register_static(&crl_apb_info);
}

type_init(crl_apb_register_types)
