/*
 * QEMU model of the Xilinx ECDSA_RSA ECDSA & RSA
 *
 * Copyright (c) 2017 Xilinx Inc.
 *
 * Partially autogenerated by xregqemu.py 2017-04-04.
 * Written by Edgar E. Iglesias.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "qemu/osdep.h"
#include "hw/sysbus.h"
#include "hw/register-dep.h"
#include "qemu/bitops.h"
#include "qemu/log.h"

#include "hw/misc/ipcores-rsa5-4k.h"

#ifndef XILINX_ECDSA_RSA_ERR_DEBUG
#define XILINX_ECDSA_RSA_ERR_DEBUG 0
#endif

#define TYPE_XILINX_ECDSA_RSA "xlnx.versal-ecdsa-rsa"

#define XILINX_ECDSA_RSA(obj) \
     OBJECT_CHECK(ECDSA_RSA, (obj), TYPE_XILINX_ECDSA_RSA)

/* Certain Windows environments incorrectly define ERROR.  */
#undef ERROR

DEP_REG32(RAM_DATA, 0x0)
DEP_REG32(RAM_ADDR, 0x4)
    DEP_FIELD(RAM_ADDR, WRRD_B, 1, 31)
    DEP_FIELD(RAM_ADDR, ADDR, 31, 0)
DEP_REG32(CTRL, 0x8)
    DEP_FIELD(CTRL, CLR_DATA_BUF, 1, 7)
    DEP_FIELD(CTRL, CLR_DONE_ABORT, 1, 6)
    DEP_FIELD(CTRL, OPCODE, 3, 0)
DEP_REG32(STATUS, 0xc)
    DEP_FIELD(STATUS, PROGRESS_CNT, 5, 3)
    DEP_FIELD(STATUS, ERROR, 1, 2)
    DEP_FIELD(STATUS, BUSY, 1, 1)
    DEP_FIELD(STATUS, DONE, 1, 0)
DEP_REG32(MINV, 0x10)
DEP_REG32(KEY_LENGTH, 0x20)
    DEP_FIELD(KEY_LENGTH, BIN_PRIME, 1, 31)
    DEP_FIELD(KEY_LENGTH, LENGTH, 15, 0)
DEP_REG32(CFG0, 0x28)
    DEP_FIELD(CFG0, QSEL, 2, 6)
    DEP_FIELD(CFG0, MULTI_PASS, 6, 0)
DEP_REG32(CFG1, 0x2c)
    DEP_FIELD(CFG1, MONT_DIGIT, 8, 0)
DEP_REG32(CFG2, 0x30)
    DEP_FIELD(CFG2, MEM_LOC_SIZE, 5, 0)
DEP_REG32(CFG3, 0x34)
    DEP_FIELD(CFG3, MONT_MOD, 4, 4)
    DEP_FIELD(CFG3, SCRATCH, 4, 0)
DEP_REG32(CFG4, 0x38)
    DEP_FIELD(CFG4, START_ADDR, 8, 0)
DEP_REG32(CFG5, 0x3c)
    DEP_FIELD(CFG5, NO_GROUPS, 5, 0)
DEP_REG32(RESET, 0x40)
    DEP_FIELD(RESET, RESET, 1, 0)
DEP_REG32(APB_SLAVE_ERR_CTRL, 0x44)
    DEP_FIELD(APB_SLAVE_ERR_CTRL, ENABLE, 1, 0)
DEP_REG32(ECDSA_RSA_ISR, 0x48)
    DEP_FIELD(ECDSA_RSA_ISR, SLVERR, 1, 1)
    DEP_FIELD(ECDSA_RSA_ISR, DONE, 1, 0)
DEP_REG32(ECDSA_RSA_IMR, 0x4c)
    DEP_FIELD(ECDSA_RSA_IMR, SLVERR, 1, 1)
    DEP_FIELD(ECDSA_RSA_IMR, DONE, 1, 0)
DEP_REG32(ECDSA_RSA_IER, 0x50)
    DEP_FIELD(ECDSA_RSA_IER, SLVERR, 1, 1)
    DEP_FIELD(ECDSA_RSA_IER, DONE, 1, 0)
DEP_REG32(ECDSA_RSA_IDR, 0x54)
    DEP_FIELD(ECDSA_RSA_IDR, SLVERR, 1, 1)
    DEP_FIELD(ECDSA_RSA_IDR, DONE, 1, 0)
DEP_REG32(RSA_CFG, 0x58)
    DEP_FIELD(RSA_CFG, RD_ENDIANNESS, 1, 1)
    DEP_FIELD(RSA_CFG, WR_ENDIANNESS, 1, 0)

#define R_MAX (R_RSA_CFG + 1)

#define R_CTRL_nop         0x00
#define R_CTRL_exp         0x01
#define R_CTRL_mod         0x02
#define R_CTRL_mul         0x03
#define R_CTRL_rrmod       0x04
#define R_CTRL_exppre      0x05
#define R_CTRL_exec        0x06
#define R_CTRL_reserved    0x07

typedef struct ECDSA_RSA {
    SysBusDevice parent_obj;
    MemoryRegion iomem;
    qemu_irq irq_ecdsa_rsa;


    IPCoresRSA rsa;
    uint8_t guard0[32 * 1024];
    /* Read-Write buffer.  */
    struct word rw_buf;
    uint8_t guard1[32 * 1024];
    uint8_t guard2[32 * 1024];

    uint32_t regs[R_MAX];
    DepRegisterInfo regs_info[R_MAX];
} ECDSA_RSA;

typedef int (*ALUFunc)(IPCoresRSA *, unsigned int, unsigned int);

static ALUFunc alu_ops[] = {
    [R_CTRL_nop] = rsa_do_nop,
    [R_CTRL_exp] = rsa_do_exp,
    [R_CTRL_mod] = rsa_do_mod,
    [R_CTRL_mul] = rsa_do_mul,
    [R_CTRL_rrmod] = rsa_do_rrmod,
    [R_CTRL_exppre] = rsa_do_exppre,
};

static void ecdsa_rsa_update_irq(ECDSA_RSA *s)
{
    bool pending;
    bool done;

    /* Propagate the DONE status bit.  */
    done = DEP_AF_EX32(s->regs, STATUS, DONE);
    DEP_AF_DP32(s->regs, ECDSA_RSA_ISR, DONE, done);

    pending = s->regs[R_ECDSA_RSA_ISR] & ~s->regs[R_ECDSA_RSA_IMR];
    qemu_set_irq(s->irq_ecdsa_rsa, pending);
}

static void ecdsa_rsa_isr_postw(DepRegisterInfo *reg, uint64_t val64)
{
    ECDSA_RSA *s = XILINX_ECDSA_RSA(reg->opaque);
    ecdsa_rsa_update_irq(s);
}

static uint64_t ecdsa_rsa_ier_prew(DepRegisterInfo *reg, uint64_t val64)
{
    ECDSA_RSA *s = XILINX_ECDSA_RSA(reg->opaque);
    uint32_t val = val64;

    s->regs[R_ECDSA_RSA_IMR] &= ~val;
    ecdsa_rsa_update_irq(s);
    return 0;
}

static uint64_t ecdsa_rsa_idr_prew(DepRegisterInfo *reg, uint64_t val64)
{
    ECDSA_RSA *s = XILINX_ECDSA_RSA(reg->opaque);
    uint32_t val = val64;

    s->regs[R_ECDSA_RSA_IMR] |= val;
    ecdsa_rsa_update_irq(s);
    return 0;
}

static void ecdsa_rsa_reset(DeviceState *dev)
{
    ECDSA_RSA *s = XILINX_ECDSA_RSA(dev);
    unsigned int i;

    for (i = 0; i < ARRAY_SIZE(s->regs_info); ++i) {
        dep_register_reset(&s->regs_info[i]);
    }

    rsa_reset(&s->rsa);
    ecdsa_rsa_update_irq(s);
}

static uint64_t ecdsa_rsa_ram_data_prew(DepRegisterInfo *reg, uint64_t val)
{
    ECDSA_RSA *s = XILINX_ECDSA_RSA(reg->opaque);
    uint32_t r = (uint32_t) val;

    if (DEP_AF_EX32(s->regs, RSA_CFG, WR_ENDIANNESS)) {
        r = bswap32(r);
    }
    return r;
}

static uint64_t ecdsa_rsa_ram_data_postr(DepRegisterInfo *reg, uint64_t val64)
{
    ECDSA_RSA *s = XILINX_ECDSA_RSA(reg->opaque);
    uint32_t r;

    r = s->rw_buf.u32[0];
    memmove(&s->rw_buf.u32[0], &s->rw_buf.u32[1],
            sizeof s->rw_buf.u32 - sizeof s->rw_buf.u32[0]);
    if (DEP_AF_EX32(s->regs, RSA_CFG, RD_ENDIANNESS)) {
        r = bswap32(r);
    }
    return r;
}

static void ecdsa_rsa_ram_data_postw(DepRegisterInfo *reg, uint64_t val64)
{
    ECDSA_RSA *s = XILINX_ECDSA_RSA(reg->opaque);

    /*
     * Six writes fill the write-buffer. LSB is written first.
     * We shift down at every write.
     */
    memmove(&s->rw_buf.u32[0], &s->rw_buf.u32[1],
            sizeof s->rw_buf.u32 - sizeof s->rw_buf.u32[0]);
    s->rw_buf.u32[ARRAY_SIZE(s->rw_buf.u32) - 1] = val64;
}

static void ecdsa_rsa_ram_addr_postw(DepRegisterInfo *reg, uint64_t val64)
{
    ECDSA_RSA *s = XILINX_ECDSA_RSA(reg->opaque);
    uint32_t val32 = val64;
    uint32_t addr = DEP_F_EX32(val32, RAM_ADDR, ADDR);

    if (DEP_F_EX32(val32, RAM_ADDR, WRRD_B)) {
        memcpy(&s->rsa.mem.words[addr].u8[0], &s->rw_buf.u8[0],
               sizeof s->rw_buf.u8);
        s->rsa.word_def[addr] = true;
    } else {
        memcpy(&s->rw_buf.u8[0], &s->rsa.mem.words[addr].u8[0],
               sizeof s->rw_buf.u8);
    }
}

static void ecdsa_rsa_run_microcode(ECDSA_RSA *s,
                                    unsigned int start_word,
                                    unsigned int digits,
                                    unsigned int loc_size,
                                    unsigned int m2_loc,
                                    bool binary)
{
    unsigned int word = start_word;
    int pc = (BITS_PER_WORD / 16) - 1;
    struct {
        unsigned int result;
        unsigned int opcode;
        unsigned int args[2];
    } insn;
    uint16_t ir;

    do {
        ir = s->rsa.mem.words[word].u16[pc];

        insn.result = extract32(ir, 0, 4);
        insn.args[1] = extract32(ir, 4, 4);
        insn.args[0] = extract32(ir, 8, 4);
        insn.opcode = extract32(ir, 12, 4);

        switch (insn.opcode) {
        case 0:
            break;
        case 1:
            if (binary) {
                rsa_do_bin_mont(&s->rsa,
                               insn.args[0] * loc_size, insn.args[1] * loc_size,
                               insn.result * loc_size, m2_loc * loc_size,
                               digits);
            } else {
                rsa_do_montmul(&s->rsa,
                               insn.args[0] * loc_size, insn.args[1] * loc_size,
                               insn.result * loc_size, m2_loc * loc_size,
                               digits);
            }
            break;
        case 2:
            if (binary) {
                rsa_do_xor(&s->rsa,
                           insn.args[0] * loc_size, insn.args[1] * loc_size,
                           insn.result * loc_size, m2_loc * loc_size,
                           digits);
            } else {
                rsa_do_add(&s->rsa,
                           insn.args[0] * loc_size, insn.args[1] * loc_size,
                           insn.result * loc_size, m2_loc * loc_size,
                           digits);
            }
            break;
        case 3:
            if (binary) {
                rsa_do_xor(&s->rsa,
                           insn.args[0] * loc_size, insn.args[1] * loc_size,
                           insn.result * loc_size, m2_loc * loc_size,
                           digits);
            } else {
                rsa_do_sub(&s->rsa,
                           insn.args[0] * loc_size, insn.args[1] * loc_size,
                           insn.result * loc_size, m2_loc * loc_size,
                           digits);
            }
            break;
        case 4:
            if (binary) {
                rsa_do_gf_mod(&s->rsa,
                              insn.args[0] * loc_size,
                              insn.args[1] * loc_size,
                              insn.result * loc_size, m2_loc * loc_size,
                              digits);
            } else {
                rsa_do_mod_addr(&s->rsa,
                                insn.args[0] * loc_size,
                                insn.args[1] * loc_size,
                                insn.result * loc_size, m2_loc * loc_size,
                                digits);
            }
            break;
        default:
            abort();
            break;
        };

        /* PC starts at top of word and counts down.  */
        pc--;
        if (pc < 0) {
            word++;
            pc = (BITS_PER_WORD / 16) - 1;
        }
    } while (insn.opcode != 0);
}

static void ecdsa_rsa_ctrl_postw(DepRegisterInfo *reg, uint64_t val64)
{
    ECDSA_RSA *s = XILINX_ECDSA_RSA(reg->opaque);
    uint32_t val32 = val64;
    unsigned int op = DEP_F_EX32(val32, CTRL, OPCODE);
    bool clr_done_abort = DEP_F_EX32(val32, CTRL, CLR_DONE_ABORT);
    bool clr_data_buf = DEP_F_EX32(val32, CTRL, CLR_DATA_BUF);
    bool binary = DEP_AF_EX32(s->regs, KEY_LENGTH, BIN_PRIME);
    unsigned int bitlen = DEP_AF_EX32(s->regs, KEY_LENGTH, LENGTH);
    unsigned int digits = DEP_AF_EX32(s->regs, CFG1, MONT_DIGIT);
    unsigned int loc_size = DEP_AF_EX32(s->regs, CFG2, MEM_LOC_SIZE);
    unsigned int start_addr = DEP_AF_EX32(s->regs, CFG4, START_ADDR);
    unsigned int m2_loc = DEP_AF_EX32(s->regs, CFG3, MONT_MOD);
    int err;

    if (clr_data_buf) {
        memset(&s->rw_buf.u8[0], 0, sizeof s->rw_buf.u8);
    }

    if (clr_done_abort) {
        DEP_AF_DP32(s->regs, STATUS, DONE, 0);
        if (op) {
            qemu_log_mask(LOG_GUEST_ERROR, "RSA: clr-done-abort op=%x\n", op);
        }
        goto done;
    }

    if (op == R_CTRL_nop) {
        goto done;
    }

    /* Clear the ERROR status for every new OP.  */
    DEP_AF_DP32(s->regs, STATUS, ERROR, 0);

    if (op <= R_CTRL_exppre) {
        err = alu_ops[op](&s->rsa, bitlen, digits);
    } else {
        ecdsa_rsa_run_microcode(s, start_addr, digits,
                                loc_size, m2_loc, binary);
        err = 0;
    }

    if (err) {
        DEP_AF_DP32(s->regs, STATUS, ERROR, 1);
        qemu_log_mask(LOG_GUEST_ERROR, "RSA: Detected an error: %s\n",
                      rsa_strerror(err));
    } else {
        DEP_AF_DP32(s->regs, STATUS, DONE, 1);
    }

done:
    ecdsa_rsa_update_irq(s);
}

static void ecdsa_rsa_minv_postw(DepRegisterInfo *reg, uint64_t val64)
{
    ECDSA_RSA *s = XILINX_ECDSA_RSA(reg->opaque);

    rsa_set_minv(&s->rsa, val64);
}

static void ecdsa_rsa_reset_postw(DepRegisterInfo *reg, uint64_t val64)
{
    ECDSA_RSA *s = XILINX_ECDSA_RSA(reg->opaque);

    ecdsa_rsa_reset(DEVICE(s));
}

static DepRegisterAccessInfo ecdsa_rsa_regs_info[] = {
    {   .name = "RAM_DATA",  .decode.addr = A_RAM_DATA,
        .post_read = ecdsa_rsa_ram_data_postr,
        .post_write = ecdsa_rsa_ram_data_postw,
        .pre_write = ecdsa_rsa_ram_data_prew,
    },{ .name = "RAM_ADDR",  .decode.addr = A_RAM_ADDR,
        .post_write = ecdsa_rsa_ram_addr_postw,
    },{ .name = "CTRL",  .decode.addr = A_CTRL,
        .rsvd = 0x38,
        .post_write = ecdsa_rsa_ctrl_postw,
    },{ .name = "STATUS",  .decode.addr = A_STATUS,
        .ro = 0xff,
    },{ .name = "MINV",  .decode.addr = A_MINV,
        .post_write = ecdsa_rsa_minv_postw,
    },{ .name = "KEY_LENGTH",  .decode.addr = A_KEY_LENGTH,
    },{ .name = "CFG0",  .decode.addr = A_CFG0,
    },{ .name = "CFG1",  .decode.addr = A_CFG1,
    },{ .name = "CFG2",  .decode.addr = A_CFG2,
    },{ .name = "CFG3",  .decode.addr = A_CFG3,
    },{ .name = "CFG4",  .decode.addr = A_CFG4,
    },{ .name = "CFG5",  .decode.addr = A_CFG5,
    },{ .name = "RESET",  .decode.addr = A_RESET,
        .reset = 0x1,
        .post_write = ecdsa_rsa_reset_postw,
    },{ .name = "APB_SLAVE_ERR_CTRL",  .decode.addr = A_APB_SLAVE_ERR_CTRL,
    },{ .name = "ECDSA_RSA_ISR",  .decode.addr = A_ECDSA_RSA_ISR,
        .w1c = 0x3,
        .post_write = ecdsa_rsa_isr_postw,
    },{ .name = "ECDSA_RSA_IMR",  .decode.addr = A_ECDSA_RSA_IMR,
        .reset = 0x3,
        .ro = 0x3,
    },{ .name = "ECDSA_RSA_IER",  .decode.addr = A_ECDSA_RSA_IER,
        .pre_write = ecdsa_rsa_ier_prew,
    },{ .name = "ECDSA_RSA_IDR",  .decode.addr = A_ECDSA_RSA_IDR,
        .pre_write = ecdsa_rsa_idr_prew,
    },{ .name = "RSA_CFG",  .decode.addr = A_RSA_CFG,
    }
};

static void check_guard(const char *name, uint8_t *g, int size)
{
    {
        int i;

        for (i = 0; i < size; i++) {
            if (g[i]) {
                qemu_log("MEM OVERWRITE! %s[%d]=%x\n", name, i, g[i]);
            }
        }
    }
}

static uint64_t ecdsa_rsa_read(void *opaque, hwaddr addr, unsigned size)
{
    ECDSA_RSA *s = XILINX_ECDSA_RSA(opaque);
    DepRegisterInfo *r = &s->regs_info[addr / 4];

    if (!r->data) {
        qemu_log("%s: Decode error: read from %" HWADDR_PRIx "\n",
                 object_get_canonical_path(OBJECT(s)),
                 addr);
        return 0;
    }
    return dep_register_read(r);
}

static void ecdsa_rsa_write(void *opaque, hwaddr addr, uint64_t value,
                      unsigned size)
{
    ECDSA_RSA *s = XILINX_ECDSA_RSA(opaque);
    DepRegisterInfo *r = &s->regs_info[addr / 4];

    if (!r->data) {
        qemu_log("%s: Decode error: write to %" HWADDR_PRIx "=%" PRIx64 "\n",
                 object_get_canonical_path(OBJECT(s)),
                 addr, value);
        return;
    }
    dep_register_write(r, value, ~0);
    check_guard("guard0", s->guard0, ARRAY_SIZE(s->guard0));
    check_guard("guard1", s->guard1, ARRAY_SIZE(s->guard1));
    check_guard("guard2", s->guard2, ARRAY_SIZE(s->guard2));
}

static const MemoryRegionOps ecdsa_rsa_ops = {
    .read = ecdsa_rsa_read,
    .write = ecdsa_rsa_write,
    .endianness = DEVICE_LITTLE_ENDIAN,
    .valid = {
        .min_access_size = 4,
        .max_access_size = 4,
    },
};

static void ecdsa_rsa_realize(DeviceState *dev, Error **errp)
{
    ECDSA_RSA *s = XILINX_ECDSA_RSA(dev);
    const char *prefix = object_get_canonical_path(OBJECT(dev));
    unsigned int i;

    for (i = 0; i < ARRAY_SIZE(ecdsa_rsa_regs_info); ++i) {
        DepRegisterInfo *r;

        r = &s->regs_info[ecdsa_rsa_regs_info[i].decode.addr / 4];
        *r = (DepRegisterInfo) {
            .data = (uint8_t *)&s->regs[
                    ecdsa_rsa_regs_info[i].decode.addr / 4],
            .data_size = sizeof(uint32_t),
            .access = &ecdsa_rsa_regs_info[i],
            .debug = XILINX_ECDSA_RSA_ERR_DEBUG,
            .prefix = prefix,
            .opaque = s,
        };
    }
}

static void ecdsa_rsa_init(Object *obj)
{
    ECDSA_RSA *s = XILINX_ECDSA_RSA(obj);
    SysBusDevice *sbd = SYS_BUS_DEVICE(obj);

    memory_region_init_io(&s->iomem, obj, &ecdsa_rsa_ops, s,
                          TYPE_XILINX_ECDSA_RSA, R_MAX * 4);
    sysbus_init_mmio(sbd, &s->iomem);
    sysbus_init_irq(sbd, &s->irq_ecdsa_rsa);
}

static const VMStateDescription vmstate_ecdsa_rsa = {
    .name = TYPE_XILINX_ECDSA_RSA,
    .version_id = 1,
    .minimum_version_id = 1,
    .minimum_version_id_old = 1,
    .fields = (VMStateField[]) {
        VMSTATE_UINT32_ARRAY(regs, ECDSA_RSA, R_MAX),
        VMSTATE_END_OF_LIST(),
    }
};

static void ecdsa_rsa_class_init(ObjectClass *klass, void *data)
{
    DeviceClass *dc = DEVICE_CLASS(klass);

    dc->reset = ecdsa_rsa_reset;
    dc->realize = ecdsa_rsa_realize;
    dc->vmsd = &vmstate_ecdsa_rsa;
}

static const TypeInfo ecdsa_rsa_info = {
    .name          = TYPE_XILINX_ECDSA_RSA,
    .parent        = TYPE_SYS_BUS_DEVICE,
    .instance_size = sizeof(ECDSA_RSA),
    .class_init    = ecdsa_rsa_class_init,
    .instance_init = ecdsa_rsa_init,
};

static void ecdsa_rsa_register_types(void)
{
    type_register_static(&ecdsa_rsa_info);
}

type_init(ecdsa_rsa_register_types)
